"""Test data loading and `dapper.xp_process.xpSpace.print`.

If script is run with one of these command-line arguments,
then there is no test generation, but rather:

- `--print` causes actual printing.
- `--replace` causes creation of a copy of this file (suffix `.new`),
  but with updated printouts in place of the `old`s.

Some notes:

- Parameterization -- A custom, hacky way to do this is used instead off
  pytest.mark.parametrize. Maybe not the best idea, but I have learned
  a fair amount doing it.
- Capturing stdout -- `xpSpace.print` only called once for each `old`
  (unlike a pytest fixture with capsys) `=>` fast.
- To better understand failing tests, use `pytest -vv` with pytest-clarity installed.
"""

import inspect
# Capture stdout
import io
import sys
from contextlib import redirect_stdout
from dataclasses import dataclass

import dapper as dpr

__pdoc__ = {}

if "--replace" in sys.argv:
    replacements = []

    @dataclass
    class _Replacement:
        lines: list
        nOpen: int
        nClose: int

else:  # Insert tests in namespace
    test_register = locals()


def backtrack_until_finding(substr, lineno):
    while True:
        lineno -= 1
        if substr in orig_code[lineno]:
            return lineno


orig_code = open(__file__, "r").readlines()

# Enable breakpoints and post-mortem debugging
_debug = "pytest" not in sys.modules and "--replace" not in sys.argv


# https://stackoverflow.com/a/22434594
def cap_stdout(fun, *args, **kwargs):
    """Capture stdout."""
    if _debug:
        fun(*args, **kwargs)
    else:
        with io.StringIO() as stringbuf, redirect_stdout(stringbuf):
            fun(*args, **kwargs)
            return stringbuf.getvalue()


def gen_test_set(xp_dict, *args, **kwargs):
    """Capture printout of `dapper.xp_process.xpSpace.print`, gen tests."""
    # Get stdout of xpSpace.print()
    output = cap_stdout(xp_dict.print, *args, **kwargs)

    # introspection
    caller_lineno = inspect.currentframe().f_back.f_lineno
    nClose = backtrack_until_finding('"""\n', caller_lineno)
    nOpen = backtrack_until_finding('"""', nClose)

    if "--print" in sys.argv:
        print(output)
        return

    if "--replace" in sys.argv:
        replacements.append(_Replacement(output.splitlines(True), nOpen, nClose))

    elif not _debug:  # Generate & register tests

        # Split string into tables
        lineno = nOpen + 1
        for table_old, table_new in zip(old.split("\n\n"), output.split("\n\n")):
            # Use kwargs to bind current values of table_old/_new
            def compare(expected=table_old, printout=table_new):
                assert printout == expected

            name = f"test_table_starting_on_line_{lineno}"
            test_register[name] = compare
            __pdoc__[name] = False

            lineno += table_old.count("\n") + 2


##
save_as = dpr.rc.dirs.DAPPER / "tests" / "data"
xps = dpr.load_xps(save_as)
xps = dpr.xpSpace.from_list(xps)

xps_shorter = dpr.xpSpace.from_list(
    [xp for xp in xps.values() if getattr(xp, "da_method") != "LETKF"])  # noqa

##
old = """Averages (in time and) over seed.

[43mTable for da_method='Climatology'[0m
 F  ⑊  rmse.a ±1σ     ☠ ✓
--  -  ------------------
 8  |  0.7775 ±0.0005 0 2
10  |  0.9728 ±0.0005 0 2

[43mTable for da_method='OptInterp'[0m
 F  ⑊  rmse.a ±1σ   ☠ ✓
--  -  ----------------
 8  |    0.14 ±0.05 0 2
10  |    0.15 ±0.05 0 2

[43mTable for da_method='Var3D'[0m
 F   xB  ⑊  rmse.a ±1σ    ☠ ✓
--  ---  -  -----------------
 8  0.1  |   0.175 ±0.004 0 2
 8  0.2  |   0.274 ±0.006 0 2
 8  0.4  |   0.392 ±0.006 0 2
10  0.1  |   0.182 ±0.006 0 2
10  0.2  |   0.282 ±0.008 0 2
10  0.4  |   0.398 ±0.007 0 2

[43mTable for da_method='EnKF'[0m
                      _______N=10_______  ________20_______  ________50_______
 F  upd_a    infl  ⑊  rmse.a ±1σ     ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓
--  -------  ----  -  ------------------  -----------------  -----------------
 8  PertObs  1.01  |  0.032  ±0.005  0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  PertObs  1.1   |  0.033  ±0.005  0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  PertObs  2     |  0.1884 ±0.0002 0 2   0.23  ±0.02  0 2   0.29  ±0.02  0 2
10  PertObs  1.01  |  0.034  ±0.005  0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2
10  PertObs  1.1   |  0.036  ±0.005  0 2   0.033 ±0.006 0 2   0.032 ±0.005 0 2
10  PertObs  2     |  0.189  ±0.004  0 2   0.24  ±0.02  0 2   0.30  ±0.02  0 2
 8  Sqrt     1.01  |  0.032  ±0.005  0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  Sqrt     1.1   |  0.033  ±0.005  0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  Sqrt     2     |  0.21   ±0.01   0 2   0.264 ±0.007 0 2   0.32  ±0.01  0 2
10  Sqrt     1.01  |  0.034  ±0.005  0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2
10  Sqrt     1.1   |  0.036  ±0.006  0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2
10  Sqrt     2     |  0.22   ±0.01   0 2   0.268 ±0.008 0 2   0.33  ±0.01  0 2

[43mTable for da_method='EnKF_N'[0m
       _______N=10______  ________20_______  ________50_______
 F  ⑊  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓
--  -  -----------------  -----------------  -----------------
 8  |   0.032 ±0.005 0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
10  |   0.034 ±0.005 0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2

[43mTable for da_method='LETKF'[0m
                      _______N=10______  ________20_______  ________50_______
 F  infl  loc_rad  ⑊  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓
--  ----  -------  -  -----------------  -----------------  -----------------
 8  1.01        1  |   0.032 ±0.005 0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  1.01        4  |   0.032 ±0.005 0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  1.01       10  |   0.032 ±0.005 0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  1.01       40  |   0.032 ±0.005 0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  1.1         1  |   0.032 ±0.005 0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  1.1         4  |   0.032 ±0.005 0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  1.1        10  |   0.033 ±0.005 0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  1.1        40  |   0.033 ±0.005 0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
 8  2           1  |   0.32  ±0.01  0 2   0.33  ±0.01  0 2   0.333 ±0.009 0 2
 8  2           4  |   0.289 ±0.006 0 2   0.31  ±0.01  0 2   0.33  ±0.01  0 2
 8  2          10  |   0.228 ±0.007 0 2   0.275 ±0.001 0 2   0.32  ±0.01  0 2
 8  2          40  |   0.21  ±0.01  0 2   0.264 ±0.007 0 2   0.32  ±0.01  0 2
10  1.01        1  |   0.034 ±0.005 0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2
10  1.01        4  |   0.034 ±0.005 0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2
10  1.01       10  |   0.034 ±0.005 0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2
10  1.01       40  |   0.034 ±0.005 0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2
10  1.1         1  |   0.034 ±0.005 0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2
10  1.1         4  |   0.034 ±0.005 0 2   0.032 ±0.005 0 2   0.032 ±0.005 0 2
10  1.1        10  |   0.035 ±0.005 0 2   0.032 ±0.005 0 2   0.032 ±0.005 0 2
10  1.1        40  |   0.035 ±0.006 0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2
10  2           1  |   0.33  ±0.01  0 2   0.330 ±0.008 0 2   0.339 ±0.008 0 2
10  2           4  |   0.294 ±0.008 0 2   0.318 ±0.009 0 2   0.33  ±0.01  0 2
10  2          10  |   0.233 ±0.006 0 2   0.281 ±0.003 0 2   0.33  ±0.01  0 2
10  2          40  |   0.22  ±0.01  0 2   0.268 ±0.008 0 2   0.33  ±0.01  0 2
"""
gen_test_set(
    xps,
    "rmse.a",
    dict(
        outer="da_method",
        inner="N",
        mean="seed",
    ),
)

##
old = """Averages (in time and) over seed.

[43mTable for da_method='Climatology'[0m
 F  ⑊  rmse.a ±1σ     *('infl',) ☠ ✓
--  -  -----------------------------
 8  |  0.7775 ±0.0005 *(None,)   0 2
10  |  0.9728 ±0.0005 *(None,)   0 2

[43mTable for da_method='OptInterp'[0m
 F  ⑊  rmse.a ±1σ   *('infl',) ☠ ✓
--  -  ---------------------------
 8  |    0.14 ±0.05 *(None,)   0 2
10  |    0.15 ±0.05 *(None,)   0 2

[43mTable for da_method='Var3D'[0m
 F   xB  ⑊  rmse.a ±1σ    *('infl',) ☠ ✓
--  ---  -  ----------------------------
 8  0.1  |   0.175 ±0.004 *(None,)   0 2
 8  0.2  |   0.274 ±0.006 *(None,)   0 2
 8  0.4  |   0.392 ±0.006 *(None,)   0 2
10  0.1  |   0.182 ±0.006 *(None,)   0 2
10  0.2  |   0.282 ±0.008 *(None,)   0 2
10  0.4  |   0.398 ±0.007 *(None,)   0 2

[43mTable for da_method='EnKF'[0m
                ____________N=10____________  _____________20_____________  _____________50_____________
 F  upd_a    ⑊  rmse.a ±1σ    *('infl',) ☠ ✓  rmse.a ±1σ    *('infl',) ☠ ✓  rmse.a ±1σ    *('infl',) ☠ ✓
--  -------  -  ----------------------------  ----------------------------  ----------------------------
 8  PertObs  |   0.032 ±0.005 *(1.01,)   0 2   0.031 ±0.005 *(1.01,)   0 2   0.030 ±0.005 *(1.1,)    0 2
10  PertObs  |   0.034 ±0.005 *(1.01,)   0 2   0.033 ±0.005 *(1.01,)   0 2   0.032 ±0.005 *(1.1,)    0 2
 8  Sqrt     |   0.032 ±0.005 *(1.01,)   0 2   0.031 ±0.005 *(1.01,)   0 2   0.030 ±0.005 *(1.1,)    0 2
10  Sqrt     |   0.034 ±0.005 *(1.01,)   0 2   0.033 ±0.005 *(1.01,)   0 2   0.032 ±0.005 *(1.1,)    0 2

[43mTable for da_method='EnKF_N'[0m
       ____________N=10____________  _____________20_____________  _____________50_____________
 F  ⑊  rmse.a ±1σ    *('infl',) ☠ ✓  rmse.a ±1σ    *('infl',) ☠ ✓  rmse.a ±1σ    *('infl',) ☠ ✓
--  -  ----------------------------  ----------------------------  ----------------------------
 8  |   0.032 ±0.005 *(1.0,)    0 2   0.031 ±0.005 *(1.0,)    0 2   0.030 ±0.005 *(1.0,)    0 2
10  |   0.034 ±0.005 *(1.0,)    0 2   0.033 ±0.005 *(1.0,)    0 2   0.032 ±0.005 *(1.0,)    0 2

[43mTable for da_method='LETKF'[0m
                ____________N=10____________  _____________20_____________  _____________50_____________
 F  loc_rad  ⑊  rmse.a ±1σ    *('infl',) ☠ ✓  rmse.a ±1σ    *('infl',) ☠ ✓  rmse.a ±1σ    *('infl',) ☠ ✓
--  -------  -  ----------------------------  ----------------------------  ----------------------------
 8        1  |   0.032 ±0.005 *(1.1,)    0 2   0.031 ±0.005 *(1.1,)    0 2   0.030 ±0.005 *(1.1,)    0 2
 8        4  |   0.032 ±0.005 *(1.01,)   0 2   0.031 ±0.005 *(1.1,)    0 2   0.030 ±0.005 *(1.1,)    0 2
 8       10  |   0.032 ±0.005 *(1.01,)   0 2   0.031 ±0.005 *(1.1,)    0 2   0.030 ±0.005 *(1.1,)    0 2
 8       40  |   0.032 ±0.005 *(1.01,)   0 2   0.031 ±0.005 *(1.01,)   0 2   0.030 ±0.005 *(1.1,)    0 2
10        1  |   0.034 ±0.005 *(1.1,)    0 2   0.033 ±0.005 *(1.1,)    0 2   0.032 ±0.005 *(1.1,)    0 2
10        4  |   0.034 ±0.005 *(1.01,)   0 2   0.032 ±0.005 *(1.1,)    0 2   0.032 ±0.005 *(1.1,)    0 2
10       10  |   0.034 ±0.005 *(1.01,)   0 2   0.032 ±0.005 *(1.1,)    0 2   0.032 ±0.005 *(1.1,)    0 2
10       40  |   0.034 ±0.005 *(1.01,)   0 2   0.033 ±0.005 *(1.01,)   0 2   0.032 ±0.005 *(1.1,)    0 2
"""
gen_test_set(xps, "rmse.a",
             dict(outer="da_method", inner="N", mean="seed", optim="infl"))

##
old = """Averages (in time and) over seed.

[43mTable for da_method='Climatology'[0m
 F  ⑊  kurt.f ±1σ  ☠ ✓
--  -  ---------------
 8  |     nan ±nan 2 0
10  |     nan ±nan 2 0

[43mTable for da_method='OptInterp'[0m
 F  ⑊  kurt.f ±1σ  ☠ ✓
--  -  ---------------
 8  |     nan ±nan 2 0
10  |     nan ±nan 2 0

[43mTable for da_method='Var3D'[0m
 F   xB  ⑊  kurt.f ±1σ  ☠ ✓
--  ---  -  ---------------
 8  0.1  |     nan ±nan 2 0
 8  0.2  |     nan ±nan 2 0
 8  0.4  |     nan ±nan 2 0
10  0.1  |     nan ±nan 2 0
10  0.2  |     nan ±nan 2 0
10  0.4  |     nan ±nan 2 0

[43mTable for da_method='EnKF'[0m
                      ________N=10_______  _________20________  ________50_______
 F  upd_a    infl  ⑊   kurt.f ±1σ     ☠ ✓   kurt.f ±1σ     ☠ ✓  kurt.f ±1σ    ☠ ✓
--  -------  ----  -  -------------------  -------------------  -----------------
 8  PertObs  1.01  |  -1.0475 ±0.0007 0 2  -0.63   ±0.01   0 2  -0.21  ±0.05  0 2
 8  PertObs  1.1   |  -1.044  ±0.006  0 2  -0.625  ±0.009  0 2  -0.21  ±0.05  0 2
 8  PertObs  2     |  -1.045  ±0.002  0 2  -0.630  ±0.007  0 2  -0.248 ±0.008 0 2
10  PertObs  1.01  |  -1.0428 ±0.0007 0 2  -0.62   ±0.01   0 2  -0.19  ±0.05  0 2
10  PertObs  1.1   |  -1.039  ±0.005  0 2  -0.62   ±0.01   0 2  -0.19  ±0.05  0 2
10  PertObs  2     |  -1.05   ±0.01   0 2  -0.61   ±0.01   0 2  -0.24  ±0.01  0 2
 8  Sqrt     1.01  |  -1.05   ±0.02   0 2  -0.6415 ±0.0004 0 2  -0.20  ±0.02  0 2
 8  Sqrt     1.1   |  -1.05   ±0.02   0 2  -0.6412 ±0.0002 0 2  -0.21  ±0.02  0 2
 8  Sqrt     2     |  -1.07   ±0.03   0 2  -0.64   ±0.02   0 2  -0.209 ±0.005 0 2
10  Sqrt     1.01  |  -1.05   ±0.02   0 2  -0.634  ±0.001  0 2  -0.19  ±0.03  0 2
10  Sqrt     1.1   |  -1.05   ±0.02   0 2  -0.634  ±0.001  0 2  -0.19  ±0.02  0 2
10  Sqrt     2     |  -1.07   ±0.02   0 2  -0.645  ±0.009  0 2  -0.208 ±0.005 0 2

[43mTable for da_method='EnKF_N'[0m
       ______N=10______  _________20________  _______50_______
 F  ⑊  kurt.f ±1σ   ☠ ✓   kurt.f ±1σ     ☠ ✓  kurt.f ±1σ   ☠ ✓
--  -  ----------------  -------------------  ----------------
 8  |   -1.05 ±0.02 0 2  -0.6415 ±0.0005 0 2   -0.20 ±0.02 0 2
10  |   -1.05 ±0.02 0 2  -0.634  ±0.001  0 2   -0.19 ±0.03 0 2

[43mTable for da_method='LETKF'[0m
                      ______N=10______  _________20________  ________50_______
 F  infl  loc_rad  ⑊  kurt.f ±1σ   ☠ ✓   kurt.f ±1σ     ☠ ✓  kurt.f ±1σ    ☠ ✓
--  ----  -------  -  ----------------  -------------------  -----------------
 8  1.01        1  |   -1.05 ±0.02 0 2  -0.6412 ±0.0003 0 2  -0.20  ±0.02  0 2
 8  1.01        4  |   -1.05 ±0.02 0 2  -0.6411 ±0.0003 0 2  -0.20  ±0.02  0 2
 8  1.01       10  |   -1.05 ±0.02 0 2  -0.6413 ±0.0004 0 2  -0.20  ±0.02  0 2
 8  1.01       40  |   -1.05 ±0.02 0 2  -0.6415 ±0.0004 0 2  -0.20  ±0.02  0 2
 8  1.1         1  |   -1.05 ±0.02 0 2  -0.6406 ±0.0001 0 2  -0.21  ±0.02  0 2
 8  1.1         4  |   -1.05 ±0.02 0 2  -0.6405 ±0.0001 0 2  -0.21  ±0.02  0 2
 8  1.1        10  |   -1.05 ±0.02 0 2  -0.6407 ±0.0001 0 2  -0.21  ±0.02  0 2
 8  1.1        40  |   -1.05 ±0.02 0 2  -0.6411 ±0.0002 0 2  -0.21  ±0.02  0 2
 8  2           1  |   -1.05 ±0.05 0 2  -0.62   ±0.03   0 2  -0.22  ±0.01  0 2
 8  2           4  |   -1.05 ±0.06 0 2  -0.62   ±0.04   0 2  -0.211 ±0.004 0 2
 8  2          10  |   -1.07 ±0.04 0 2  -0.63   ±0.02   0 2  -0.209 ±0.005 0 2
 8  2          40  |   -1.07 ±0.03 0 2  -0.64   ±0.02   0 2  -0.209 ±0.005 0 2
10  1.01        1  |   -1.05 ±0.02 0 2  -0.633  ±0.002  0 2  -0.19  ±0.03  0 2
10  1.01        4  |   -1.05 ±0.02 0 2  -0.633  ±0.002  0 2  -0.19  ±0.03  0 2
10  1.01       10  |   -1.05 ±0.02 0 2  -0.633  ±0.001  0 2  -0.19  ±0.03  0 2
10  1.01       40  |   -1.05 ±0.02 0 2  -0.634  ±0.001  0 2  -0.19  ±0.03  0 2
10  1.1         1  |   -1.05 ±0.02 0 2  -0.633  ±0.002  0 2  -0.19  ±0.03  0 2
10  1.1         4  |   -1.05 ±0.02 0 2  -0.633  ±0.002  0 2  -0.19  ±0.02  0 2
10  1.1        10  |   -1.05 ±0.02 0 2  -0.633  ±0.002  0 2  -0.19  ±0.02  0 2
10  1.1        40  |   -1.05 ±0.02 0 2  -0.634  ±0.001  0 2  -0.19  ±0.02  0 2
10  2           1  |   -1.04 ±0.03 0 2  -0.61   ±0.02   0 2  -0.21  ±0.02  0 2
10  2           4  |   -1.04 ±0.04 0 2  -0.61   ±0.02   0 2  -0.210 ±0.006 0 2
10  2          10  |   -1.07 ±0.03 0 2  -0.632  ±0.003  0 2  -0.209 ±0.004 0 2
10  2          40  |   -1.07 ±0.02 0 2  -0.644  ±0.009  0 2  -0.208 ±0.004 0 2
"""
gen_test_set(
    xps,
    "kurt.f",
    dict(
        outer="da_method",
        inner="N",
        mean="seed",
    ),
)

##
old = """Averages (in time and) over seed.

[43mTable for da_method='Climatology'[0m
 F  ⑊  rmse.a ±1σ     ☠ ✓
--  -  ------------------
 8  |  0.7775 ±0.0005 0 2
10  |  0.9728 ±0.0005 0 2

[43mTable for da_method='OptInterp'[0m
 F  ⑊  rmse.a ±1σ   ☠ ✓
--  -  ----------------
 8  |    0.14 ±0.05 0 2
10  |    0.15 ±0.05 0 2

[43mTable for da_method='Var3D'[0m
 F   xB  ⑊  rmse.a ±1σ    ☠ ✓
--  ---  -  -----------------
 8  0.1  |   0.175 ±0.004 0 2
 8  0.2  |   0.274 ±0.006 0 2
 8  0.4  |   0.392 ±0.006 0 2
10  0.1  |   0.182 ±0.006 0 2
10  0.2  |   0.282 ±0.008 0 2
10  0.4  |   0.398 ±0.007 0 2

[43mTable for da_method='EnKF'[0m
                    ____infl=1.01____  _______1.1_______  _______2.0________
 F  upd_a     N  ⑊  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ     ☠ ✓
--  -------  --  -  -----------------  -----------------  ------------------
 8  PertObs  10  |   0.032 ±0.005 0 2   0.033 ±0.005 0 2  0.1884 ±0.0002 0 2
 8  PertObs  20  |   0.031 ±0.005 0 2   0.031 ±0.005 0 2  0.23   ±0.02   0 2
 8  PertObs  50  |   0.030 ±0.005 0 2   0.030 ±0.005 0 2  0.29   ±0.02   0 2
10  PertObs  10  |   0.034 ±0.005 0 2   0.036 ±0.005 0 2  0.189  ±0.004  0 2
10  PertObs  20  |   0.033 ±0.005 0 2   0.033 ±0.006 0 2  0.24   ±0.02   0 2
10  PertObs  50  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2  0.30   ±0.02   0 2
 8  Sqrt     10  |   0.032 ±0.005 0 2   0.033 ±0.005 0 2  0.21   ±0.01   0 2
 8  Sqrt     20  |   0.031 ±0.005 0 2   0.031 ±0.005 0 2  0.264  ±0.007  0 2
 8  Sqrt     50  |   0.030 ±0.005 0 2   0.030 ±0.005 0 2  0.32   ±0.01   0 2
10  Sqrt     10  |   0.034 ±0.005 0 2   0.036 ±0.006 0 2  0.22   ±0.01   0 2
10  Sqrt     20  |   0.033 ±0.005 0 2   0.033 ±0.005 0 2  0.268  ±0.008  0 2
10  Sqrt     50  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2  0.33   ±0.01   0 2

[43mTable for da_method='EnKF_N'[0m
 F   N  ⑊  rmse.a ±1σ    ☠ ✓
--  --  -  -----------------
 8  10  |   0.032 ±0.005 0 2
 8  20  |   0.031 ±0.005 0 2
 8  50  |   0.030 ±0.005 0 2
10  10  |   0.034 ±0.005 0 2
10  20  |   0.033 ±0.005 0 2
10  50  |   0.032 ±0.005 0 2

[43mTable for da_method='LETKF'[0m
                    ____infl=1.01____  _______1.1_______  _______2.0_______
 F   N  loc_rad  ⑊  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓
--  --  -------  -  -----------------  -----------------  -----------------
 8  10        1  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.32  ±0.01  0 2
 8  10        4  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.289 ±0.006 0 2
 8  10       10  |   0.032 ±0.005 0 2   0.033 ±0.005 0 2   0.228 ±0.007 0 2
 8  10       40  |   0.032 ±0.005 0 2   0.033 ±0.005 0 2   0.21  ±0.01  0 2
 8  20        1  |   0.031 ±0.005 0 2   0.031 ±0.005 0 2   0.33  ±0.01  0 2
 8  20        4  |   0.031 ±0.005 0 2   0.031 ±0.005 0 2   0.31  ±0.01  0 2
 8  20       10  |   0.031 ±0.005 0 2   0.031 ±0.005 0 2   0.275 ±0.001 0 2
 8  20       40  |   0.031 ±0.005 0 2   0.031 ±0.005 0 2   0.264 ±0.007 0 2
 8  50        1  |   0.030 ±0.005 0 2   0.030 ±0.005 0 2   0.333 ±0.009 0 2
 8  50        4  |   0.030 ±0.005 0 2   0.030 ±0.005 0 2   0.33  ±0.01  0 2
 8  50       10  |   0.030 ±0.005 0 2   0.030 ±0.005 0 2   0.32  ±0.01  0 2
 8  50       40  |   0.030 ±0.005 0 2   0.030 ±0.005 0 2   0.32  ±0.01  0 2
10  10        1  |   0.034 ±0.005 0 2   0.034 ±0.005 0 2   0.33  ±0.01  0 2
10  10        4  |   0.034 ±0.005 0 2   0.034 ±0.005 0 2   0.294 ±0.008 0 2
10  10       10  |   0.034 ±0.005 0 2   0.035 ±0.005 0 2   0.233 ±0.006 0 2
10  10       40  |   0.034 ±0.005 0 2   0.035 ±0.006 0 2   0.22  ±0.01  0 2
10  20        1  |   0.033 ±0.005 0 2   0.033 ±0.005 0 2   0.330 ±0.008 0 2
10  20        4  |   0.033 ±0.005 0 2   0.032 ±0.005 0 2   0.318 ±0.009 0 2
10  20       10  |   0.033 ±0.005 0 2   0.032 ±0.005 0 2   0.281 ±0.003 0 2
10  20       40  |   0.033 ±0.005 0 2   0.033 ±0.005 0 2   0.268 ±0.008 0 2
10  50        1  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.339 ±0.008 0 2
10  50        4  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.33  ±0.01  0 2
10  50       10  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.33  ±0.01  0 2
10  50       40  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.33  ±0.01  0 2
"""
gen_test_set(
    xps,
    "rmse.a",
    dict(
        outer="da_method",
        inner="infl",
        mean="seed",
    ),
)

##
old = """Averages (in time and) over seed.

[43mTable for N=None[0m
da_method     F   xB  ⑊  rmse.a ±1σ     ☠ ✓
-----------  --  ---  -  ------------------
Climatology   8       |  0.7775 ±0.0005 0 2
Climatology  10       |  0.9728 ±0.0005 0 2
OptInterp     8       |  0.14   ±0.05   0 2
OptInterp    10       |  0.15   ±0.05   0 2
Var3D         8  0.1  |  0.175  ±0.004  0 2
Var3D         8  0.2  |  0.274  ±0.006  0 2
Var3D         8  0.4  |  0.392  ±0.006  0 2
Var3D        10  0.1  |  0.182  ±0.006  0 2
Var3D        10  0.2  |  0.282  ±0.008  0 2
Var3D        10  0.4  |  0.398  ±0.007  0 2

[43mTable for N=10[0m
                                    ____infl=1.01____  _______1.1_______  _______2.0________  _______1.0_______
da_method   F  upd_a    loc_rad  ⑊  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ     ☠ ✓  rmse.a ±1σ    ☠ ✓
---------  --  -------  -------  -  -----------------  -----------------  ------------------  -----------------
EnKF        8  PertObs           |   0.032 ±0.005 0 2   0.033 ±0.005 0 2  0.1884 ±0.0002 0 2         ±         
EnKF       10  PertObs           |   0.034 ±0.005 0 2   0.036 ±0.005 0 2  0.189  ±0.004  0 2         ±         
EnKF        8  Sqrt              |   0.032 ±0.005 0 2   0.033 ±0.005 0 2  0.21   ±0.01   0 2         ±         
EnKF       10  Sqrt              |   0.034 ±0.005 0 2   0.036 ±0.006 0 2  0.22   ±0.01   0 2         ±         
EnKF_N      8                    |         ±                  ±                  ±             0.032 ±0.005 0 2
EnKF_N     10                    |         ±                  ±                  ±             0.034 ±0.005 0 2
LETKF       8                 1  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2  0.32   ±0.01   0 2         ±         
LETKF       8                 4  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2  0.289  ±0.006  0 2         ±         
LETKF       8                10  |   0.032 ±0.005 0 2   0.033 ±0.005 0 2  0.228  ±0.007  0 2         ±         
LETKF       8                40  |   0.032 ±0.005 0 2   0.033 ±0.005 0 2  0.21   ±0.01   0 2         ±         
LETKF      10                 1  |   0.034 ±0.005 0 2   0.034 ±0.005 0 2  0.33   ±0.01   0 2         ±         
LETKF      10                 4  |   0.034 ±0.005 0 2   0.034 ±0.005 0 2  0.294  ±0.008  0 2         ±         
LETKF      10                10  |   0.034 ±0.005 0 2   0.035 ±0.005 0 2  0.233  ±0.006  0 2         ±         
LETKF      10                40  |   0.034 ±0.005 0 2   0.035 ±0.006 0 2  0.22   ±0.01   0 2         ±         

[43mTable for N=20[0m
                                    ____infl=1.01____  _______1.1_______  _______2.0_______  _______1.0_______
da_method   F  upd_a    loc_rad  ⑊  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓
---------  --  -------  -------  -  -----------------  -----------------  -----------------  -----------------
EnKF        8  PertObs           |   0.031 ±0.005 0 2   0.031 ±0.005 0 2   0.23  ±0.02  0 2         ±         
EnKF       10  PertObs           |   0.033 ±0.005 0 2   0.033 ±0.006 0 2   0.24  ±0.02  0 2         ±         
EnKF        8  Sqrt              |   0.031 ±0.005 0 2   0.031 ±0.005 0 2   0.264 ±0.007 0 2         ±         
EnKF       10  Sqrt              |   0.033 ±0.005 0 2   0.033 ±0.005 0 2   0.268 ±0.008 0 2         ±         
EnKF_N      8                    |         ±                  ±                  ±            0.031 ±0.005 0 2
EnKF_N     10                    |         ±                  ±                  ±            0.033 ±0.005 0 2
LETKF       8                 1  |   0.031 ±0.005 0 2   0.031 ±0.005 0 2   0.33  ±0.01  0 2         ±         
LETKF       8                 4  |   0.031 ±0.005 0 2   0.031 ±0.005 0 2   0.31  ±0.01  0 2         ±         
LETKF       8                10  |   0.031 ±0.005 0 2   0.031 ±0.005 0 2   0.275 ±0.001 0 2         ±         
LETKF       8                40  |   0.031 ±0.005 0 2   0.031 ±0.005 0 2   0.264 ±0.007 0 2         ±         
LETKF      10                 1  |   0.033 ±0.005 0 2   0.033 ±0.005 0 2   0.330 ±0.008 0 2         ±         
LETKF      10                 4  |   0.033 ±0.005 0 2   0.032 ±0.005 0 2   0.318 ±0.009 0 2         ±         
LETKF      10                10  |   0.033 ±0.005 0 2   0.032 ±0.005 0 2   0.281 ±0.003 0 2         ±         
LETKF      10                40  |   0.033 ±0.005 0 2   0.033 ±0.005 0 2   0.268 ±0.008 0 2         ±         

[43mTable for N=50[0m
                                    ____infl=1.01____  _______1.1_______  _______2.0_______  _______1.0_______
da_method   F  upd_a    loc_rad  ⑊  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓
---------  --  -------  -------  -  -----------------  -----------------  -----------------  -----------------
EnKF        8  PertObs           |   0.030 ±0.005 0 2   0.030 ±0.005 0 2   0.29  ±0.02  0 2         ±         
EnKF       10  PertObs           |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.30  ±0.02  0 2         ±         
EnKF        8  Sqrt              |   0.030 ±0.005 0 2   0.030 ±0.005 0 2   0.32  ±0.01  0 2         ±         
EnKF       10  Sqrt              |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.33  ±0.01  0 2         ±         
EnKF_N      8                    |         ±                  ±                  ±            0.030 ±0.005 0 2
EnKF_N     10                    |         ±                  ±                  ±            0.032 ±0.005 0 2
LETKF       8                 1  |   0.030 ±0.005 0 2   0.030 ±0.005 0 2   0.333 ±0.009 0 2         ±         
LETKF       8                 4  |   0.030 ±0.005 0 2   0.030 ±0.005 0 2   0.33  ±0.01  0 2         ±         
LETKF       8                10  |   0.030 ±0.005 0 2   0.030 ±0.005 0 2   0.32  ±0.01  0 2         ±         
LETKF       8                40  |   0.030 ±0.005 0 2   0.030 ±0.005 0 2   0.32  ±0.01  0 2         ±         
LETKF      10                 1  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.339 ±0.008 0 2         ±         
LETKF      10                 4  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.33  ±0.01  0 2         ±         
LETKF      10                10  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.33  ±0.01  0 2         ±         
LETKF      10                40  |   0.032 ±0.005 0 2   0.032 ±0.005 0 2   0.33  ±0.01  0 2         ±         
"""
gen_test_set(
    xps,
    "rmse.a",
    dict(
        outer="N",
        inner="infl",
        mean="seed",
    ),
)

##
old = """Averages (in time and) over seed.

[43mTable for N=None[0m
            da_method='Climatology'  ___OptInterp____  ______Var3D______
 F   xB  ⑊  rmse.a ±1σ     ☠ ✓       rmse.a ±1σ   ☠ ✓  rmse.a ±1σ    ☠ ✓
--  ---  -  -----------------------  ----------------  -----------------
 8       |  0.7775 ±0.0005 0 2         0.14 ±0.05 0 2         ±         
10       |  0.9728 ±0.0005 0 2         0.15 ±0.05 0 2         ±         
 8  0.1  |         ±                        ±           0.175 ±0.004 0 2
 8  0.2  |         ±                        ±           0.274 ±0.006 0 2
 8  0.4  |         ±                        ±           0.392 ±0.006 0 2
10  0.1  |         ±                        ±           0.182 ±0.006 0 2
10  0.2  |         ±                        ±           0.282 ±0.008 0 2
10  0.4  |         ±                        ±           0.398 ±0.007 0 2

[43mTable for N=10[0m
                               _da_method='EnKF'_  ______EnKF_N_____  ______LETKF______
 F  upd_a    infl  loc_rad  ⑊  rmse.a ±1σ     ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓
--  -------  ----  -------  -  ------------------  -----------------  -----------------
 8  PertObs  1.01           |  0.032  ±0.005  0 2         ±                  ±         
 8  PertObs  1.1            |  0.033  ±0.005  0 2         ±                  ±         
 8  PertObs  2              |  0.1884 ±0.0002 0 2         ±                  ±         
10  PertObs  1.01           |  0.034  ±0.005  0 2         ±                  ±         
10  PertObs  1.1            |  0.036  ±0.005  0 2         ±                  ±         
10  PertObs  2              |  0.189  ±0.004  0 2         ±                  ±         
 8  Sqrt     1.01           |  0.032  ±0.005  0 2         ±                  ±         
 8  Sqrt     1.1            |  0.033  ±0.005  0 2         ±                  ±         
 8  Sqrt     2              |  0.21   ±0.01   0 2         ±                  ±         
10  Sqrt     1.01           |  0.034  ±0.005  0 2         ±                  ±         
10  Sqrt     1.1            |  0.036  ±0.006  0 2         ±                  ±         
10  Sqrt     2              |  0.22   ±0.01   0 2         ±                  ±         
 8           1              |         ±             0.032 ±0.005 0 2         ±         
10           1              |         ±             0.034 ±0.005 0 2         ±         
 8           1.01        1  |         ±                   ±            0.032 ±0.005 0 2
 8           1.01        4  |         ±                   ±            0.032 ±0.005 0 2
 8           1.01       10  |         ±                   ±            0.032 ±0.005 0 2
 8           1.01       40  |         ±                   ±            0.032 ±0.005 0 2
 8           1.1         1  |         ±                   ±            0.032 ±0.005 0 2
 8           1.1         4  |         ±                   ±            0.032 ±0.005 0 2
 8           1.1        10  |         ±                   ±            0.033 ±0.005 0 2
 8           1.1        40  |         ±                   ±            0.033 ±0.005 0 2
 8           2           1  |         ±                   ±            0.32  ±0.01  0 2
 8           2           4  |         ±                   ±            0.289 ±0.006 0 2
 8           2          10  |         ±                   ±            0.228 ±0.007 0 2
 8           2          40  |         ±                   ±            0.21  ±0.01  0 2
10           1.01        1  |         ±                   ±            0.034 ±0.005 0 2
10           1.01        4  |         ±                   ±            0.034 ±0.005 0 2
10           1.01       10  |         ±                   ±            0.034 ±0.005 0 2
10           1.01       40  |         ±                   ±            0.034 ±0.005 0 2
10           1.1         1  |         ±                   ±            0.034 ±0.005 0 2
10           1.1         4  |         ±                   ±            0.034 ±0.005 0 2
10           1.1        10  |         ±                   ±            0.035 ±0.005 0 2
10           1.1        40  |         ±                   ±            0.035 ±0.006 0 2
10           2           1  |         ±                   ±            0.33  ±0.01  0 2
10           2           4  |         ±                   ±            0.294 ±0.008 0 2
10           2          10  |         ±                   ±            0.233 ±0.006 0 2
10           2          40  |         ±                   ±            0.22  ±0.01  0 2

[43mTable for N=20[0m
                               _da_method='EnKF'  ______EnKF_N_____  ______LETKF______
 F  upd_a    infl  loc_rad  ⑊  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓
--  -------  ----  -------  -  -----------------  -----------------  -----------------
 8  PertObs  1.01           |   0.031 ±0.005 0 2         ±                  ±         
 8  PertObs  1.1            |   0.031 ±0.005 0 2         ±                  ±         
 8  PertObs  2              |   0.23  ±0.02  0 2         ±                  ±         
10  PertObs  1.01           |   0.033 ±0.005 0 2         ±                  ±         
10  PertObs  1.1            |   0.033 ±0.006 0 2         ±                  ±         
10  PertObs  2              |   0.24  ±0.02  0 2         ±                  ±         
 8  Sqrt     1.01           |   0.031 ±0.005 0 2         ±                  ±         
 8  Sqrt     1.1            |   0.031 ±0.005 0 2         ±                  ±         
 8  Sqrt     2              |   0.264 ±0.007 0 2         ±                  ±         
10  Sqrt     1.01           |   0.033 ±0.005 0 2         ±                  ±         
10  Sqrt     1.1            |   0.033 ±0.005 0 2         ±                  ±         
10  Sqrt     2              |   0.268 ±0.008 0 2         ±                  ±         
 8           1              |         ±            0.031 ±0.005 0 2         ±         
10           1              |         ±            0.033 ±0.005 0 2         ±         
 8           1.01        1  |         ±                  ±            0.031 ±0.005 0 2
 8           1.01        4  |         ±                  ±            0.031 ±0.005 0 2
 8           1.01       10  |         ±                  ±            0.031 ±0.005 0 2
 8           1.01       40  |         ±                  ±            0.031 ±0.005 0 2
 8           1.1         1  |         ±                  ±            0.031 ±0.005 0 2
 8           1.1         4  |         ±                  ±            0.031 ±0.005 0 2
 8           1.1        10  |         ±                  ±            0.031 ±0.005 0 2
 8           1.1        40  |         ±                  ±            0.031 ±0.005 0 2
 8           2           1  |         ±                  ±            0.33  ±0.01  0 2
 8           2           4  |         ±                  ±            0.31  ±0.01  0 2
 8           2          10  |         ±                  ±            0.275 ±0.001 0 2
 8           2          40  |         ±                  ±            0.264 ±0.007 0 2
10           1.01        1  |         ±                  ±            0.033 ±0.005 0 2
10           1.01        4  |         ±                  ±            0.033 ±0.005 0 2
10           1.01       10  |         ±                  ±            0.033 ±0.005 0 2
10           1.01       40  |         ±                  ±            0.033 ±0.005 0 2
10           1.1         1  |         ±                  ±            0.033 ±0.005 0 2
10           1.1         4  |         ±                  ±            0.032 ±0.005 0 2
10           1.1        10  |         ±                  ±            0.032 ±0.005 0 2
10           1.1        40  |         ±                  ±            0.033 ±0.005 0 2
10           2           1  |         ±                  ±            0.330 ±0.008 0 2
10           2           4  |         ±                  ±            0.318 ±0.009 0 2
10           2          10  |         ±                  ±            0.281 ±0.003 0 2
10           2          40  |         ±                  ±            0.268 ±0.008 0 2

[43mTable for N=50[0m
                               _da_method='EnKF'  ______EnKF_N_____  ______LETKF______
 F  upd_a    infl  loc_rad  ⑊  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓
--  -------  ----  -------  -  -----------------  -----------------  -----------------
 8  PertObs  1.01           |   0.030 ±0.005 0 2         ±                  ±         
 8  PertObs  1.1            |   0.030 ±0.005 0 2         ±                  ±         
 8  PertObs  2              |   0.29  ±0.02  0 2         ±                  ±         
10  PertObs  1.01           |   0.032 ±0.005 0 2         ±                  ±         
10  PertObs  1.1            |   0.032 ±0.005 0 2         ±                  ±         
10  PertObs  2              |   0.30  ±0.02  0 2         ±                  ±         
 8  Sqrt     1.01           |   0.030 ±0.005 0 2         ±                  ±         
 8  Sqrt     1.1            |   0.030 ±0.005 0 2         ±                  ±         
 8  Sqrt     2              |   0.32  ±0.01  0 2         ±                  ±         
10  Sqrt     1.01           |   0.032 ±0.005 0 2         ±                  ±         
10  Sqrt     1.1            |   0.032 ±0.005 0 2         ±                  ±         
10  Sqrt     2              |   0.33  ±0.01  0 2         ±                  ±         
 8           1              |         ±            0.030 ±0.005 0 2         ±         
10           1              |         ±            0.032 ±0.005 0 2         ±         
 8           1.01        1  |         ±                  ±            0.030 ±0.005 0 2
 8           1.01        4  |         ±                  ±            0.030 ±0.005 0 2
 8           1.01       10  |         ±                  ±            0.030 ±0.005 0 2
 8           1.01       40  |         ±                  ±            0.030 ±0.005 0 2
 8           1.1         1  |         ±                  ±            0.030 ±0.005 0 2
 8           1.1         4  |         ±                  ±            0.030 ±0.005 0 2
 8           1.1        10  |         ±                  ±            0.030 ±0.005 0 2
 8           1.1        40  |         ±                  ±            0.030 ±0.005 0 2
 8           2           1  |         ±                  ±            0.333 ±0.009 0 2
 8           2           4  |         ±                  ±            0.33  ±0.01  0 2
 8           2          10  |         ±                  ±            0.32  ±0.01  0 2
 8           2          40  |         ±                  ±            0.32  ±0.01  0 2
10           1.01        1  |         ±                  ±            0.032 ±0.005 0 2
10           1.01        4  |         ±                  ±            0.032 ±0.005 0 2
10           1.01       10  |         ±                  ±            0.032 ±0.005 0 2
10           1.01       40  |         ±                  ±            0.032 ±0.005 0 2
10           1.1         1  |         ±                  ±            0.032 ±0.005 0 2
10           1.1         4  |         ±                  ±            0.032 ±0.005 0 2
10           1.1        10  |         ±                  ±            0.032 ±0.005 0 2
10           1.1        40  |         ±                  ±            0.032 ±0.005 0 2
10           2           1  |         ±                  ±            0.339 ±0.008 0 2
10           2           4  |         ±                  ±            0.33  ±0.01  0 2
10           2          10  |         ±                  ±            0.33  ±0.01  0 2
10           2          40  |         ±                  ±            0.33  ±0.01  0 2
"""
gen_test_set(
    xps,
    "rmse.a",
    dict(
        outer="N",
        inner="da_method",
        mean="seed",
    ),
)

##
old = """Averages in time only (=> the 1σ estimates may be unreliable).

[43mTable for N=None[0m
                         _seed=3000__  ____3001____
da_method     F   xB  ⑊  rmse.a ±1σ    rmse.a ±1σ  
-----------  --  ---  -  ------------  ------------
Climatology   8       |    0.8  ±0.2     0.8  ±0.2 
Climatology  10       |    1.0  ±0.3     1.0  ±0.3 
OptInterp     8       |    0.09 ±0.03    0.19 ±0.03
OptInterp    10       |    0.10 ±0.02    0.20 ±0.03
Var3D         8  0.1  |    0.18 ±0.02    0.17 ±0.02
Var3D         8  0.2  |    0.28 ±0.02    0.27 ±0.02
Var3D         8  0.4  |    0.40 ±0.02    0.39 ±0.01
Var3D        10  0.1  |    0.19 ±0.03    0.18 ±0.02
Var3D        10  0.2  |    0.29 ±0.03    0.27 ±0.02
Var3D        10  0.4  |    0.40 ±0.02    0.39 ±0.01

[43mTable for N=10[0m
                                          __seed=3000___  _____3001_____
da_method   F  upd_a    infl  loc_rad  ⑊  rmse.a ±1σ      rmse.a ±1σ    
---------  --  -------  ----  -------  -  --------------  --------------
EnKF        8  PertObs  1.01           |  0.0370 ±0.0009  0.028  ±0.001 
EnKF        8  PertObs  1.1            |  0.038  ±0.002   0.029  ±0.001 
EnKF        8  PertObs  2              |  0.19   ±0.08    0.19   ±0.06  
EnKF       10  PertObs  1.01           |  0.039  ±0.002   0.0291 ±0.0009
EnKF       10  PertObs  1.1            |  0.041  ±0.004   0.031  ±0.002 
EnKF       10  PertObs  2              |  0.18   ±0.08    0.19   ±0.07  
EnKF        8  Sqrt     1.01           |  0.0372 ±0.0009  0.028  ±0.001 
EnKF        8  Sqrt     1.1            |  0.038  ±0.002   0.028  ±0.001 
EnKF        8  Sqrt     2              |  0.2    ±0.1     0.20   ±0.08  
EnKF       10  Sqrt     1.01           |  0.040  ±0.002   0.0290 ±0.0009
EnKF       10  Sqrt     1.1            |  0.041  ±0.004   0.030  ±0.002 
EnKF       10  Sqrt     2              |  0.2    ±0.1     0.20   ±0.08  
EnKF_N      8           1              |  0.0371 ±0.0009  0.028  ±0.001 
EnKF_N     10           1              |  0.039  ±0.002   0.0290 ±0.0009
LETKF       8           1.01        1  |  0.0369 ±0.0009  0.0277 ±0.0009
LETKF       8           1.01        4  |  0.0369 ±0.0008  0.028  ±0.001 
LETKF       8           1.01       10  |  0.0369 ±0.0009  0.028  ±0.001 
LETKF       8           1.01       40  |  0.0371 ±0.0009  0.028  ±0.001 
LETKF       8           1.1         1  |  0.0367 ±0.0009  0.0277 ±0.0009
LETKF       8           1.1         4  |  0.0371 ±0.0009  0.0276 ±0.0009
LETKF       8           1.1        10  |  0.037  ±0.001   0.028  ±0.001 
LETKF       8           1.1        40  |  0.038  ±0.002   0.028  ±0.001 
LETKF       8           2           1  |  0.3    ±0.2     0.3    ±0.2   
LETKF       8           2           4  |  0.3    ±0.1     0.3    ±0.1   
LETKF       8           2          10  |  0.2    ±0.1     0.22   ±0.09  
LETKF       8           2          40  |  0.2    ±0.1     0.20   ±0.08  
LETKF      10           1.01        1  |  0.039  ±0.002   0.0291 ±0.0009
LETKF      10           1.01        4  |  0.039  ±0.002   0.0289 ±0.0008
LETKF      10           1.01       10  |  0.039  ±0.002   0.0289 ±0.0008
LETKF      10           1.01       40  |  0.039  ±0.002   0.0290 ±0.0009
LETKF      10           1.1         1  |  0.039  ±0.002   0.0292 ±0.0009
LETKF      10           1.1         4  |  0.040  ±0.002   0.029  ±0.001 
LETKF      10           1.1        10  |  0.040  ±0.003   0.029  ±0.001 
LETKF      10           1.1        40  |  0.041  ±0.004   0.030  ±0.002 
LETKF      10           2           1  |  0.3    ±0.2     0.3    ±0.2   
LETKF      10           2           4  |  0.3    ±0.1     0.3    ±0.1   
LETKF      10           2          10  |  0.2    ±0.1     0.2    ±0.1   
LETKF      10           2          40  |  0.2    ±0.1     0.20   ±0.08  

[43mTable for N=20[0m
                                          __seed=3000___  _____3001_____
da_method   F  upd_a    infl  loc_rad  ⑊  rmse.a ±1σ      rmse.a ±1σ    
---------  --  -------  ----  -------  -  --------------  --------------
EnKF        8  PertObs  1.01           |  0.0359 ±0.0009  0.0260 ±0.0008
EnKF        8  PertObs  1.1            |  0.036  ±0.001   0.0261 ±0.0009
EnKF        8  PertObs  2              |  0.3    ±0.1     0.21   ±0.09  
EnKF       10  PertObs  1.01           |  0.038  ±0.002   0.0274 ±0.0009
EnKF       10  PertObs  1.1            |  0.039  ±0.002   0.028  ±0.001 
EnKF       10  PertObs  2              |  0.3    ±0.1     0.22   ±0.09  
EnKF        8  Sqrt     1.01           |  0.0360 ±0.0008  0.0260 ±0.0008
EnKF        8  Sqrt     1.1            |  0.036  ±0.001   0.0261 ±0.0009
EnKF        8  Sqrt     2              |  0.3    ±0.1     0.3    ±0.1   
EnKF       10  Sqrt     1.01           |  0.038  ±0.002   0.0274 ±0.0009
EnKF       10  Sqrt     1.1            |  0.038  ±0.002   0.028  ±0.001 
EnKF       10  Sqrt     2              |  0.3    ±0.1     0.3    ±0.1   
EnKF_N      8           1              |  0.0360 ±0.0008  0.0261 ±0.0008
EnKF_N     10           1              |  0.038  ±0.002   0.0275 ±0.0009
LETKF       8           1.01        1  |  0.0359 ±0.0008  0.0262 ±0.0007
LETKF       8           1.01        4  |  0.0357 ±0.0008  0.0261 ±0.0007
LETKF       8           1.01       10  |  0.0357 ±0.0009  0.0260 ±0.0008
LETKF       8           1.01       40  |  0.0359 ±0.0008  0.0260 ±0.0008
LETKF       8           1.1         1  |  0.0358 ±0.0008  0.0262 ±0.0007
LETKF       8           1.1         4  |  0.0356 ±0.0008  0.0259 ±0.0008
LETKF       8           1.1        10  |  0.0356 ±0.0009  0.0259 ±0.0009
LETKF       8           1.1        40  |  0.0362 ±0.0009  0.0261 ±0.0009
LETKF       8           2           1  |  0.3    ±0.2     0.3    ±0.2   
LETKF       8           2           4  |  0.3    ±0.1     0.3    ±0.2   
LETKF       8           2          10  |  0.3    ±0.1     0.3    ±0.1   
LETKF       8           2          40  |  0.3    ±0.1     0.3    ±0.1   
LETKF      10           1.01        1  |  0.038  ±0.002   0.028  ±0.001 
LETKF      10           1.01        4  |  0.038  ±0.001   0.0275 ±0.0009
LETKF      10           1.01       10  |  0.038  ±0.001   0.0274 ±0.0009
LETKF      10           1.01       40  |  0.038  ±0.002   0.0274 ±0.0009
LETKF      10           1.1         1  |  0.038  ±0.002   0.0276 ±0.0009
LETKF      10           1.1         4  |  0.038  ±0.001   0.027  ±0.001 
LETKF      10           1.1        10  |  0.038  ±0.001   0.0274 ±0.0008
LETKF      10           1.1        40  |  0.038  ±0.002   0.028  ±0.001 
LETKF      10           2           1  |  0.3    ±0.2     0.3    ±0.2   
LETKF      10           2           4  |  0.3    ±0.1     0.3    ±0.2   
LETKF      10           2          10  |  0.3    ±0.1     0.3    ±0.1   
LETKF      10           2          40  |  0.3    ±0.1     0.3    ±0.1   

[43mTable for N=50[0m
                                          __seed=3000___  _____3001_____
da_method   F  upd_a    infl  loc_rad  ⑊  rmse.a ±1σ      rmse.a ±1σ    
---------  --  -------  ----  -------  -  --------------  --------------
EnKF        8  PertObs  1.01           |  0.0352 ±0.0009  0.0253 ±0.0008
EnKF        8  PertObs  1.1            |  0.0350 ±0.0009  0.0253 ±0.0008
EnKF        8  PertObs  2              |  0.3    ±0.1     0.3    ±0.1   
EnKF       10  PertObs  1.01           |  0.037  ±0.001   0.0266 ±0.0008
EnKF       10  PertObs  1.1            |  0.037  ±0.001   0.027  ±0.001 
EnKF       10  PertObs  2              |  0.3    ±0.1     0.3    ±0.2   
EnKF        8  Sqrt     1.01           |  0.0352 ±0.0009  0.0253 ±0.0008
EnKF        8  Sqrt     1.1            |  0.0350 ±0.0009  0.0253 ±0.0008
EnKF        8  Sqrt     2              |  0.3    ±0.2     0.3    ±0.2   
EnKF       10  Sqrt     1.01           |  0.037  ±0.001   0.0266 ±0.0008
EnKF       10  Sqrt     1.1            |  0.037  ±0.001   0.027  ±0.001 
EnKF       10  Sqrt     2              |  0.3    ±0.2     0.3    ±0.2   
EnKF_N      8           1              |  0.0352 ±0.0009  0.0253 ±0.0008
EnKF_N     10           1              |  0.037  ±0.002   0.0266 ±0.0008
LETKF       8           1.01        1  |  0.0352 ±0.0008  0.0250 ±0.0008
LETKF       8           1.01        4  |  0.0350 ±0.0009  0.0250 ±0.0008
LETKF       8           1.01       10  |  0.0349 ±0.0009  0.0251 ±0.0008
LETKF       8           1.01       40  |  0.0351 ±0.0009  0.0253 ±0.0008
LETKF       8           1.1         1  |  0.0350 ±0.0009  0.0249 ±0.0008
LETKF       8           1.1         4  |  0.0347 ±0.0009  0.0249 ±0.0008
LETKF       8           1.1        10  |  0.0346 ±0.0009  0.0250 ±0.0008
LETKF       8           1.1        40  |  0.0350 ±0.0009  0.0253 ±0.0008
LETKF       8           2           1  |  0.3    ±0.2     0.3    ±0.2   
LETKF       8           2           4  |  0.3    ±0.2     0.3    ±0.2   
LETKF       8           2          10  |  0.3    ±0.2     0.3    ±0.2   
LETKF       8           2          40  |  0.3    ±0.2     0.3    ±0.2   
LETKF      10           1.01        1  |  0.037  ±0.002   0.0264 ±0.0008
LETKF      10           1.01        4  |  0.037  ±0.002   0.0264 ±0.0008
LETKF      10           1.01       10  |  0.037  ±0.001   0.0265 ±0.0008
LETKF      10           1.01       40  |  0.037  ±0.001   0.0266 ±0.0008
LETKF      10           1.1         1  |  0.037  ±0.002   0.0263 ±0.0008
LETKF      10           1.1         4  |  0.037  ±0.001   0.026  ±0.001 
LETKF      10           1.1        10  |  0.036  ±0.001   0.027  ±0.001 
LETKF      10           1.1        40  |  0.037  ±0.001   0.027  ±0.001 
LETKF      10           2           1  |  0.3    ±0.2     0.3    ±0.2   
LETKF      10           2           4  |  0.3    ±0.2     0.3    ±0.2   
LETKF      10           2          10  |  0.3    ±0.2     0.3    ±0.2   
LETKF      10           2          40  |  0.3    ±0.2     0.3    ±0.2   
"""
gen_test_set(xps, "rmse.a", dict(outer="N", inner="seed"))

##
old = """Averages in time only (=> the 1σ estimates may be unreliable).

[43mTable for da_method='Climatology'[0m
seed   F  ⑊  rmse.a ±1σ 
----  --  -  -----------
3000   8  |     0.8 ±0.2
3000  10  |     1.0 ±0.3
3001   8  |     0.8 ±0.2
3001  10  |     1.0 ±0.3

[43mTable for da_method='OptInterp'[0m
seed   F  ⑊  rmse.a ±1σ  
----  --  -  ------------
3000   8  |    0.09 ±0.03
3000  10  |    0.10 ±0.02
3001   8  |    0.19 ±0.03
3001  10  |    0.20 ±0.03

[43mTable for da_method='Var3D'[0m
seed   F   xB  ⑊  rmse.a ±1σ  
----  --  ---  -  ------------
3000   8  0.1  |    0.18 ±0.02
3000   8  0.2  |    0.28 ±0.02
3000   8  0.4  |    0.40 ±0.02
3000  10  0.1  |    0.19 ±0.03
3000  10  0.2  |    0.29 ±0.03
3000  10  0.4  |    0.40 ±0.02
3001   8  0.1  |    0.17 ±0.02
3001   8  0.2  |    0.27 ±0.02
3001   8  0.4  |    0.39 ±0.01
3001  10  0.1  |    0.18 ±0.02
3001  10  0.2  |    0.27 ±0.02
3001  10  0.4  |    0.39 ±0.01

[43mTable for da_method='EnKF'[0m
                            _____N=10_____  ______20______  ______50______
seed   F  upd_a    infl  ⑊  rmse.a ±1σ      rmse.a ±1σ      rmse.a ±1σ    
----  --  -------  ----  -  --------------  --------------  --------------
3000   8  PertObs  1.01  |  0.0370 ±0.0009  0.0359 ±0.0009  0.0352 ±0.0009
3000   8  PertObs  1.1   |  0.038  ±0.002   0.036  ±0.001   0.0350 ±0.0009
3000   8  PertObs  2     |  0.19   ±0.08    0.3    ±0.1     0.3    ±0.1   
3000  10  PertObs  1.01  |  0.039  ±0.002   0.038  ±0.002   0.037  ±0.001 
3000  10  PertObs  1.1   |  0.041  ±0.004   0.039  ±0.002   0.037  ±0.001 
3000  10  PertObs  2     |  0.18   ±0.08    0.3    ±0.1     0.3    ±0.1   
3001   8  PertObs  1.01  |  0.028  ±0.001   0.0260 ±0.0008  0.0253 ±0.0008
3001   8  PertObs  1.1   |  0.029  ±0.001   0.0261 ±0.0009  0.0253 ±0.0008
3001   8  PertObs  2     |  0.19   ±0.06    0.21   ±0.09    0.3    ±0.1   
3001  10  PertObs  1.01  |  0.0291 ±0.0009  0.0274 ±0.0009  0.0266 ±0.0008
3001  10  PertObs  1.1   |  0.031  ±0.002   0.028  ±0.001   0.027  ±0.001 
3001  10  PertObs  2     |  0.19   ±0.07    0.22   ±0.09    0.3    ±0.2   
3000   8  Sqrt     1.01  |  0.0372 ±0.0009  0.0360 ±0.0008  0.0352 ±0.0009
3000   8  Sqrt     1.1   |  0.038  ±0.002   0.036  ±0.001   0.0350 ±0.0009
3000   8  Sqrt     2     |  0.2    ±0.1     0.3    ±0.1     0.3    ±0.2   
3000  10  Sqrt     1.01  |  0.040  ±0.002   0.038  ±0.002   0.037  ±0.001 
3000  10  Sqrt     1.1   |  0.041  ±0.004   0.038  ±0.002   0.037  ±0.001 
3000  10  Sqrt     2     |  0.2    ±0.1     0.3    ±0.1     0.3    ±0.2   
3001   8  Sqrt     1.01  |  0.028  ±0.001   0.0260 ±0.0008  0.0253 ±0.0008
3001   8  Sqrt     1.1   |  0.028  ±0.001   0.0261 ±0.0009  0.0253 ±0.0008
3001   8  Sqrt     2     |  0.20   ±0.08    0.3    ±0.1     0.3    ±0.2   
3001  10  Sqrt     1.01  |  0.0290 ±0.0009  0.0274 ±0.0009  0.0266 ±0.0008
3001  10  Sqrt     1.1   |  0.030  ±0.002   0.028  ±0.001   0.027  ±0.001 
3001  10  Sqrt     2     |  0.20   ±0.08    0.3    ±0.1     0.3    ±0.2   

[43mTable for da_method='EnKF_N'[0m
             _____N=10_____  ______20______  ______50______
seed   F  ⑊  rmse.a ±1σ      rmse.a ±1σ      rmse.a ±1σ    
----  --  -  --------------  --------------  --------------
3000   8  |  0.0371 ±0.0009  0.0360 ±0.0008  0.0352 ±0.0009
3000  10  |  0.039  ±0.002   0.038  ±0.002   0.037  ±0.002 
3001   8  |  0.028  ±0.001   0.0261 ±0.0008  0.0253 ±0.0008
3001  10  |  0.0290 ±0.0009  0.0275 ±0.0009  0.0266 ±0.0008
"""
gen_test_set(xps_shorter, "rmse.a", dict(outer="da_method", inner="N"))

##
old = """Averages in time only (=> the 1σ estimates may be unreliable).

[43mTable for da_method='Climatology'[0m
seed   F  ⑊  rmse.a ±1σ  ☠ ✓
----  --  -  ---------------
3000   8  |  0.777  ±nan 0 1
3000  10  |  0.9723 ±nan 0 1
3001   8  |  0.778  ±nan 0 1
3001  10  |  0.9734 ±nan 0 1

[43mTable for da_method='OptInterp'[0m
seed   F  ⑊   rmse.a ±1σ  ☠ ✓
----  --  -  ----------------
3000   8  |  0.09469 ±nan 0 1
3000  10  |  0.09856 ±nan 0 1
3001   8  |  0.1924  ±nan 0 1
3001  10  |  0.1954  ±nan 0 1

[43mTable for da_method='Var3D'[0m
seed   F   xB  ⑊  rmse.a ±1σ  ☠ ✓
----  --  ---  -  ---------------
3000   8  0.1  |  0.1794 ±nan 0 1
3000   8  0.2  |  0.2807 ±nan 0 1
3000   8  0.4  |  0.3975 ±nan 0 1
3000  10  0.1  |  0.1877 ±nan 0 1
3000  10  0.2  |  0.2902 ±nan 0 1
3000  10  0.4  |  0.4049 ±nan 0 1
3001   8  0.1  |  0.1711 ±nan 0 1
3001   8  0.2  |  0.2681 ±nan 0 1
3001   8  0.4  |  0.3864 ±nan 0 1
3001  10  0.1  |  0.1767 ±nan 0 1
3001  10  0.2  |  0.2737 ±nan 0 1
3001  10  0.4  |  0.3902 ±nan 0 1

[43mTable for da_method='EnKF'[0m
                            ______N=10______  _______20_______  _______50_______
seed   F  upd_a    infl  ⑊   rmse.a ±1σ  ☠ ✓   rmse.a ±1σ  ☠ ✓   rmse.a ±1σ  ☠ ✓
----  --  -------  ----  -  ----------------  ----------------  ----------------
3000   8  PertObs  1.01  |  0.03704 ±nan 0 1  0.03594 ±nan 0 1  0.03519 ±nan 0 1
3000   8  PertObs  1.1   |  0.0384  ±nan 0 1  0.03638 ±nan 0 1  0.03503 ±nan 0 1
3000   8  PertObs  2     |  0.1885  ±nan 0 1  0.2527  ±nan 0 1  0.2733  ±nan 0 1
3000  10  PertObs  1.01  |  0.03936 ±nan 0 1  0.03801 ±nan 0 1  0.03722 ±nan 0 1
3000  10  PertObs  1.1   |  0.04113 ±nan 0 1  0.0386  ±nan 0 1  0.03691 ±nan 0 1
3000  10  PertObs  2     |  0.1848  ±nan 0 1  0.2508  ±nan 0 1  0.2817  ±nan 0 1
3001   8  PertObs  1.01  |  0.02768 ±nan 0 1  0.02604 ±nan 0 1  0.02528 ±nan 0 1
3001   8  PertObs  1.1   |  0.02854 ±nan 0 1  0.02608 ±nan 0 1  0.02534 ±nan 0 1
3001   8  PertObs  2     |  0.1882  ±nan 0 1  0.2115  ±nan 0 1  0.3144  ±nan 0 1
3001  10  PertObs  1.01  |  0.02908 ±nan 0 1  0.02741 ±nan 0 1  0.02664 ±nan 0 1
3001  10  PertObs  1.1   |  0.03085 ±nan 0 1  0.02756 ±nan 0 1  0.02694 ±nan 0 1
3001  10  PertObs  2     |  0.1933  ±nan 0 1  0.2194  ±nan 0 1  0.3229  ±nan 0 1
3000   8  Sqrt     1.01  |  0.03717 ±nan 0 1  0.03597 ±nan 0 1  0.03517 ±nan 0 1
3000   8  Sqrt     1.1   |  0.03847 ±nan 0 1  0.0363  ±nan 0 1  0.03505 ±nan 0 1
3000   8  Sqrt     2     |  0.2272  ±nan 0 1  0.2711  ±nan 0 1  0.3084  ±nan 0 1
3000  10  Sqrt     1.01  |  0.03953 ±nan 0 1  0.03805 ±nan 0 1  0.03719 ±nan 0 1
3000  10  Sqrt     1.1   |  0.04116 ±nan 0 1  0.03839 ±nan 0 1  0.03692 ±nan 0 1
3000  10  Sqrt     2     |  0.2296  ±nan 0 1  0.2767  ±nan 0 1  0.3141  ±nan 0 1
3001   8  Sqrt     1.01  |  0.02767 ±nan 0 1  0.02605 ±nan 0 1  0.02528 ±nan 0 1
3001   8  Sqrt     1.1   |  0.02813 ±nan 0 1  0.0261  ±nan 0 1  0.02531 ±nan 0 1
3001   8  Sqrt     2     |  0.201   ±nan 0 1  0.2575  ±nan 0 1  0.3376  ±nan 0 1
3001  10  Sqrt     1.01  |  0.02904 ±nan 0 1  0.02744 ±nan 0 1  0.02663 ±nan 0 1
3001  10  Sqrt     1.1   |  0.03004 ±nan 0 1  0.02766 ±nan 0 1  0.02685 ±nan 0 1
3001  10  Sqrt     2     |  0.2043  ±nan 0 1  0.2599  ±nan 0 1  0.343   ±nan 0 1

[43mTable for da_method='EnKF_N'[0m
             ______N=10______  _______20_______  _______50_______
seed   F  ⑊   rmse.a ±1σ  ☠ ✓   rmse.a ±1σ  ☠ ✓   rmse.a ±1σ  ☠ ✓
----  --  -  ----------------  ----------------  ----------------
3000   8  |  0.03713 ±nan 0 1  0.03597 ±nan 0 1  0.0352  ±nan 0 1
3000  10  |  0.03947 ±nan 0 1  0.03805 ±nan 0 1  0.03724 ±nan 0 1
3001   8  |  0.02768 ±nan 0 1  0.02607 ±nan 0 1  0.02529 ±nan 0 1
3001  10  |  0.02904 ±nan 0 1  0.02746 ±nan 0 1  0.02663 ±nan 0 1
"""
gen_test_set(xps_shorter, "rmse.a", dict(outer="da_method", inner="N", mean=()))

##
old = """Averages (in time and) over ('seed', 'infl').

[43mTable for da_method='Climatology'[0m
 F  ⑊  rmse.a ±1σ     ☠ ✓
--  -  ------------------
 8  |  0.7775 ±0.0005 0 2
10  |  0.9728 ±0.0005 0 2

[43mTable for da_method='OptInterp'[0m
 F  ⑊  rmse.a ±1σ   ☠ ✓
--  -  ----------------
 8  |    0.14 ±0.05 0 2
10  |    0.15 ±0.05 0 2

[43mTable for da_method='Var3D'[0m
 F   xB  ⑊  rmse.a ±1σ    ☠ ✓
--  ---  -  -----------------
 8  0.1  |   0.175 ±0.004 0 2
 8  0.2  |   0.274 ±0.006 0 2
 8  0.4  |   0.392 ±0.006 0 2
10  0.1  |   0.182 ±0.006 0 2
10  0.2  |   0.282 ±0.008 0 2
10  0.4  |   0.398 ±0.007 0 2

[43mTable for da_method='EnKF'[0m
                ______N=10______  _______20_______  _______50_______
 F  upd_a    ⑊  rmse.a ±1σ   ☠ ✓  rmse.a ±1σ   ☠ ✓  rmse.a ±1σ   ☠ ✓
--  -------  -  ----------------  ----------------  ----------------
 8  PertObs  |    0.08 ±0.03 0 6    0.10 ±0.04 0 6    0.12 ±0.06 0 6
10  PertObs  |    0.09 ±0.03 0 6    0.10 ±0.04 0 6    0.12 ±0.06 0 6
 8  Sqrt     |    0.09 ±0.04 0 6    0.11 ±0.05 0 6    0.13 ±0.06 0 6
10  Sqrt     |    0.10 ±0.04 0 6    0.11 ±0.05 0 6    0.13 ±0.06 0 6

[43mTable for da_method='EnKF_N'[0m
       _______N=10______  ________20_______  ________50_______
 F  ⑊  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓  rmse.a ±1σ    ☠ ✓
--  -  -----------------  -----------------  -----------------
 8  |   0.032 ±0.005 0 2   0.031 ±0.005 0 2   0.030 ±0.005 0 2
10  |   0.034 ±0.005 0 2   0.033 ±0.005 0 2   0.032 ±0.005 0 2
"""
gen_test_set(xps_shorter, "rmse.a",
             dict(outer="da_method", inner="N", mean=("seed", "infl")))

##
old = """Averages (in time and) over seed.

[43mTable for da_method='Climatology'[0m
 F  ⑊  rmse.a ±1σ     ☠ ✓
--  -  ------------------
 8  |  0.7775 ±0.0005 0 2
10  |  0.9728 ±0.0005 0 2

[43mTable for da_method='OptInterp'[0m
 F  ⑊  rmse.a ±1σ   ☠ ✓
--  -  ----------------
 8  |    0.14 ±0.05 0 2
10  |    0.15 ±0.05 0 2

[43mTable for da_method='Var3D'[0m
 F   xB  ⑊  rmse.a ±1σ    ☠ ✓
--  ---  -  -----------------
 8  0.1  |   0.175 ±0.004 0 2
 8  0.2  |   0.274 ±0.006 0 2
 8  0.4  |   0.392 ±0.006 0 2
10  0.1  |   0.182 ±0.006 0 2
10  0.2  |   0.282 ±0.008 0 2
10  0.4  |   0.398 ±0.007 0 2

[43mTable for da_method='EnKF'[0m
 F  upd_a     N  infl  ⑊  rmse.a ±1σ     ☠ ✓
--  -------  --  ----  -  ------------------
 8  PertObs  10  1.01  |  0.032  ±0.005  0 2
 8  PertObs  10  1.1   |  0.033  ±0.005  0 2
 8  PertObs  10  2     |  0.1884 ±0.0002 0 2
 8  PertObs  20  1.01  |  0.031  ±0.005  0 2
 8  PertObs  20  1.1   |  0.031  ±0.005  0 2
 8  PertObs  20  2     |  0.23   ±0.02   0 2
 8  PertObs  50  1.01  |  0.030  ±0.005  0 2
 8  PertObs  50  1.1   |  0.030  ±0.005  0 2
 8  PertObs  50  2     |  0.29   ±0.02   0 2
10  PertObs  10  1.01  |  0.034  ±0.005  0 2
10  PertObs  10  1.1   |  0.036  ±0.005  0 2
10  PertObs  10  2     |  0.189  ±0.004  0 2
10  PertObs  20  1.01  |  0.033  ±0.005  0 2
10  PertObs  20  1.1   |  0.033  ±0.006  0 2
10  PertObs  20  2     |  0.24   ±0.02   0 2
10  PertObs  50  1.01  |  0.032  ±0.005  0 2
10  PertObs  50  1.1   |  0.032  ±0.005  0 2
10  PertObs  50  2     |  0.30   ±0.02   0 2
 8  Sqrt     10  1.01  |  0.032  ±0.005  0 2
 8  Sqrt     10  1.1   |  0.033  ±0.005  0 2
 8  Sqrt     10  2     |  0.21   ±0.01   0 2
 8  Sqrt     20  1.01  |  0.031  ±0.005  0 2
 8  Sqrt     20  1.1   |  0.031  ±0.005  0 2
 8  Sqrt     20  2     |  0.264  ±0.007  0 2
 8  Sqrt     50  1.01  |  0.030  ±0.005  0 2
 8  Sqrt     50  1.1   |  0.030  ±0.005  0 2
 8  Sqrt     50  2     |  0.32   ±0.01   0 2
10  Sqrt     10  1.01  |  0.034  ±0.005  0 2
10  Sqrt     10  1.1   |  0.036  ±0.006  0 2
10  Sqrt     10  2     |  0.22   ±0.01   0 2
10  Sqrt     20  1.01  |  0.033  ±0.005  0 2
10  Sqrt     20  1.1   |  0.033  ±0.005  0 2
10  Sqrt     20  2     |  0.268  ±0.008  0 2
10  Sqrt     50  1.01  |  0.032  ±0.005  0 2
10  Sqrt     50  1.1   |  0.032  ±0.005  0 2
10  Sqrt     50  2     |  0.33   ±0.01   0 2

[43mTable for da_method='EnKF_N'[0m
 F   N  ⑊  rmse.a ±1σ    ☠ ✓
--  --  -  -----------------
 8  10  |   0.032 ±0.005 0 2
 8  20  |   0.031 ±0.005 0 2
 8  50  |   0.030 ±0.005 0 2
10  10  |   0.034 ±0.005 0 2
10  20  |   0.033 ±0.005 0 2
10  50  |   0.032 ±0.005 0 2
"""
gen_test_set(xps_shorter, "rmse.a", dict(outer="da_method", mean=("seed")))

##
old = """Averages in time only (=> the 1σ estimates may be unreliable).

[43mTable for da_method='Climatology'[0m
seed   F  ⑊  rmse.a ±1σ 
----  --  -  -----------
3000   8  |     0.8 ±0.2
3000  10  |     1.0 ±0.3
3001   8  |     0.8 ±0.2
3001  10  |     1.0 ±0.3

[43mTable for da_method='OptInterp'[0m
seed   F  ⑊  rmse.a ±1σ  
----  --  -  ------------
3000   8  |    0.09 ±0.03
3000  10  |    0.10 ±0.02
3001   8  |    0.19 ±0.03
3001  10  |    0.20 ±0.03

[43mTable for da_method='Var3D'[0m
seed   F   xB  ⑊  rmse.a ±1σ  
----  --  ---  -  ------------
3000   8  0.1  |    0.18 ±0.02
3000   8  0.2  |    0.28 ±0.02
3000   8  0.4  |    0.40 ±0.02
3000  10  0.1  |    0.19 ±0.03
3000  10  0.2  |    0.29 ±0.03
3000  10  0.4  |    0.40 ±0.02
3001   8  0.1  |    0.17 ±0.02
3001   8  0.2  |    0.27 ±0.02
3001   8  0.4  |    0.39 ±0.01
3001  10  0.1  |    0.18 ±0.02
3001  10  0.2  |    0.27 ±0.02
3001  10  0.4  |    0.39 ±0.01

[43mTable for da_method='EnKF'[0m
seed   F  upd_a     N  infl  ⑊  rmse.a ±1σ    
----  --  -------  --  ----  -  --------------
3000   8  PertObs  10  1.01  |  0.0370 ±0.0009
3000   8  PertObs  10  1.1   |  0.038  ±0.002 
3000   8  PertObs  10  2     |  0.19   ±0.08  
3000   8  PertObs  20  1.01  |  0.0359 ±0.0009
3000   8  PertObs  20  1.1   |  0.036  ±0.001 
3000   8  PertObs  20  2     |  0.3    ±0.1   
3000   8  PertObs  50  1.01  |  0.0352 ±0.0009
3000   8  PertObs  50  1.1   |  0.0350 ±0.0009
3000   8  PertObs  50  2     |  0.3    ±0.1   
3000  10  PertObs  10  1.01  |  0.039  ±0.002 
3000  10  PertObs  10  1.1   |  0.041  ±0.004 
3000  10  PertObs  10  2     |  0.18   ±0.08  
3000  10  PertObs  20  1.01  |  0.038  ±0.002 
3000  10  PertObs  20  1.1   |  0.039  ±0.002 
3000  10  PertObs  20  2     |  0.3    ±0.1   
3000  10  PertObs  50  1.01  |  0.037  ±0.001 
3000  10  PertObs  50  1.1   |  0.037  ±0.001 
3000  10  PertObs  50  2     |  0.3    ±0.1   
3001   8  PertObs  10  1.01  |  0.028  ±0.001 
3001   8  PertObs  10  1.1   |  0.029  ±0.001 
3001   8  PertObs  10  2     |  0.19   ±0.06  
3001   8  PertObs  20  1.01  |  0.0260 ±0.0008
3001   8  PertObs  20  1.1   |  0.0261 ±0.0009
3001   8  PertObs  20  2     |  0.21   ±0.09  
3001   8  PertObs  50  1.01  |  0.0253 ±0.0008
3001   8  PertObs  50  1.1   |  0.0253 ±0.0008
3001   8  PertObs  50  2     |  0.3    ±0.1   
3001  10  PertObs  10  1.01  |  0.0291 ±0.0009
3001  10  PertObs  10  1.1   |  0.031  ±0.002 
3001  10  PertObs  10  2     |  0.19   ±0.07  
3001  10  PertObs  20  1.01  |  0.0274 ±0.0009
3001  10  PertObs  20  1.1   |  0.028  ±0.001 
3001  10  PertObs  20  2     |  0.22   ±0.09  
3001  10  PertObs  50  1.01  |  0.0266 ±0.0008
3001  10  PertObs  50  1.1   |  0.027  ±0.001 
3001  10  PertObs  50  2     |  0.3    ±0.2   
3000   8  Sqrt     10  1.01  |  0.0372 ±0.0009
3000   8  Sqrt     10  1.1   |  0.038  ±0.002 
3000   8  Sqrt     10  2     |  0.2    ±0.1   
3000   8  Sqrt     20  1.01  |  0.0360 ±0.0008
3000   8  Sqrt     20  1.1   |  0.036  ±0.001 
3000   8  Sqrt     20  2     |  0.3    ±0.1   
3000   8  Sqrt     50  1.01  |  0.0352 ±0.0009
3000   8  Sqrt     50  1.1   |  0.0350 ±0.0009
3000   8  Sqrt     50  2     |  0.3    ±0.2   
3000  10  Sqrt     10  1.01  |  0.040  ±0.002 
3000  10  Sqrt     10  1.1   |  0.041  ±0.004 
3000  10  Sqrt     10  2     |  0.2    ±0.1   
3000  10  Sqrt     20  1.01  |  0.038  ±0.002 
3000  10  Sqrt     20  1.1   |  0.038  ±0.002 
3000  10  Sqrt     20  2     |  0.3    ±0.1   
3000  10  Sqrt     50  1.01  |  0.037  ±0.001 
3000  10  Sqrt     50  1.1   |  0.037  ±0.001 
3000  10  Sqrt     50  2     |  0.3    ±0.2   
3001   8  Sqrt     10  1.01  |  0.028  ±0.001 
3001   8  Sqrt     10  1.1   |  0.028  ±0.001 
3001   8  Sqrt     10  2     |  0.20   ±0.08  
3001   8  Sqrt     20  1.01  |  0.0260 ±0.0008
3001   8  Sqrt     20  1.1   |  0.0261 ±0.0009
3001   8  Sqrt     20  2     |  0.3    ±0.1   
3001   8  Sqrt     50  1.01  |  0.0253 ±0.0008
3001   8  Sqrt     50  1.1   |  0.0253 ±0.0008
3001   8  Sqrt     50  2     |  0.3    ±0.2   
3001  10  Sqrt     10  1.01  |  0.0290 ±0.0009
3001  10  Sqrt     10  1.1   |  0.030  ±0.002 
3001  10  Sqrt     10  2     |  0.20   ±0.08  
3001  10  Sqrt     20  1.01  |  0.0274 ±0.0009
3001  10  Sqrt     20  1.1   |  0.028  ±0.001 
3001  10  Sqrt     20  2     |  0.3    ±0.1   
3001  10  Sqrt     50  1.01  |  0.0266 ±0.0008
3001  10  Sqrt     50  1.1   |  0.027  ±0.001 
3001  10  Sqrt     50  2     |  0.3    ±0.2   

[43mTable for da_method='EnKF_N'[0m
seed   F   N  ⑊  rmse.a ±1σ    
----  --  --  -  --------------
3000   8  10  |  0.0371 ±0.0009
3000   8  20  |  0.0360 ±0.0008
3000   8  50  |  0.0352 ±0.0009
3000  10  10  |  0.039  ±0.002 
3000  10  20  |  0.038  ±0.002 
3000  10  50  |  0.037  ±0.002 
3001   8  10  |  0.028  ±0.001 
3001   8  20  |  0.0261 ±0.0008
3001   8  50  |  0.0253 ±0.0008
3001  10  10  |  0.0290 ±0.0009
3001  10  20  |  0.0275 ±0.0009
3001  10  50  |  0.0266 ±0.0008
"""
gen_test_set(xps_shorter, "rmse.a", dict(outer="da_method"))

##
old = """Averages (in time and) over seed.

                           da_method='Climatology',N=None  OptInterp,None  Var3D,None  EnKF,10  EnKF,20  EnKF,50  EnKF_N,10  EnKF_N,20  EnKF_N,50
 F   xB  upd_a    infl  ⑊  rmse.a                          rmse.a          rmse.a      rmse.a   rmse.a   rmse.a   rmse.a     rmse.a     rmse.a
--  ---  -------  ----  -  ------------------------------  --------------  ----------  -------  -------  -------  ---------  ---------  ---------
 8                      |  0.7775                            0.14                                                                             
10                      |  0.9728                            0.15                                                                             
 8  0.1                 |                                                   0.175                                                             
 8  0.2                 |                                                   0.274                                                             
 8  0.4                 |                                                   0.392                                                             
10  0.1                 |                                                   0.182                                                             
10  0.2                 |                                                   0.282                                                             
10  0.4                 |                                                   0.398                                                             
 8       PertObs  1.01  |                                                              0.032     0.031    0.030                               
 8       PertObs  1.1   |                                                              0.033     0.031    0.030                               
 8       PertObs  2     |                                                              0.1884    0.23     0.29                                
10       PertObs  1.01  |                                                              0.034     0.033    0.032                               
10       PertObs  1.1   |                                                              0.036     0.033    0.032                               
10       PertObs  2     |                                                              0.189     0.24     0.30                                
 8       Sqrt     1.01  |                                                              0.032     0.031    0.030                               
 8       Sqrt     1.1   |                                                              0.033     0.031    0.030                               
 8       Sqrt     2     |                                                              0.21      0.264    0.32                                
10       Sqrt     1.01  |                                                              0.034     0.033    0.032                               
10       Sqrt     1.1   |                                                              0.036     0.033    0.032                               
10       Sqrt     2     |                                                              0.22      0.268    0.33                                
 8                1     |                                                                                          0.032      0.031      0.030
10                1     |                                                                                          0.034      0.033      0.032
"""
gen_test_set(xps_shorter,
             "rmse.a",
             dict(inner=("da_method", "N"), mean="seed"),
             subcols=False)

##
old = """Averages in time only (=> the 1σ estimates may be unreliable).

[43mTable for da_method='Climatology',seed=3000[0m
 F  ⑊  rmse.a
--  -  ------
 8  |     0.8
10  |     1.0

[43mTable for da_method='Climatology',seed=3001[0m
 F  ⑊  rmse.a
--  -  ------
 8  |     0.8
10  |     1.0

[43mTable for da_method='OptInterp',seed=3000[0m
 F  ⑊  rmse.a
--  -  ------
 8  |    0.09
10  |    0.10

[43mTable for da_method='OptInterp',seed=3001[0m
 F  ⑊  rmse.a
--  -  ------
 8  |    0.19
10  |    0.20

[43mTable for da_method='Var3D',seed=3000[0m
 F   xB  ⑊  rmse.a
--  ---  -  ------
 8  0.1  |    0.18
 8  0.2  |    0.28
 8  0.4  |    0.40
10  0.1  |    0.19
10  0.2  |    0.29
10  0.4  |    0.40

[43mTable for da_method='Var3D',seed=3001[0m
 F   xB  ⑊  rmse.a
--  ---  -  ------
 8  0.1  |    0.17
 8  0.2  |    0.27
 8  0.4  |    0.39
10  0.1  |    0.18
10  0.2  |    0.27
10  0.4  |    0.39

[43mTable for da_method='EnKF',seed=3000[0m
                      _N=10_  __20__  __50__
 F  upd_a    infl  ⑊  rmse.a  rmse.a  rmse.a
--  -------  ----  -  ------  ------  ------
 8  PertObs  1.01  |  0.0370  0.0359  0.0352
 8  PertObs  1.1   |  0.038   0.036   0.0350
 8  PertObs  2     |  0.19    0.3     0.3   
10  PertObs  1.01  |  0.039   0.038   0.037 
10  PertObs  1.1   |  0.041   0.039   0.037 
10  PertObs  2     |  0.18    0.3     0.3   
 8  Sqrt     1.01  |  0.0372  0.0360  0.0352
 8  Sqrt     1.1   |  0.038   0.036   0.0350
 8  Sqrt     2     |  0.2     0.3     0.3   
10  Sqrt     1.01  |  0.040   0.038   0.037 
10  Sqrt     1.1   |  0.041   0.038   0.037 
10  Sqrt     2     |  0.2     0.3     0.3   

[43mTable for da_method='EnKF',seed=3001[0m
                      _N=10_  __20__  __50__
 F  upd_a    infl  ⑊  rmse.a  rmse.a  rmse.a
--  -------  ----  -  ------  ------  ------
 8  PertObs  1.01  |  0.028   0.0260  0.0253
 8  PertObs  1.1   |  0.029   0.0261  0.0253
 8  PertObs  2     |  0.19    0.21    0.3   
10  PertObs  1.01  |  0.0291  0.0274  0.0266
10  PertObs  1.1   |  0.031   0.028   0.027 
10  PertObs  2     |  0.19    0.22    0.3   
 8  Sqrt     1.01  |  0.028   0.0260  0.0253
 8  Sqrt     1.1   |  0.028   0.0261  0.0253
 8  Sqrt     2     |  0.20    0.3     0.3   
10  Sqrt     1.01  |  0.0290  0.0274  0.0266
10  Sqrt     1.1   |  0.030   0.028   0.027 
10  Sqrt     2     |  0.20    0.3     0.3   

[43mTable for da_method='EnKF_N',seed=3000[0m
       _N=10_  __20__  __50__
 F  ⑊  rmse.a  rmse.a  rmse.a
--  -  ------  ------  ------
 8  |  0.0371  0.0360  0.0352
10  |  0.039   0.038   0.037 

[43mTable for da_method='EnKF_N',seed=3001[0m
       _N=10_  __20__  __50__
 F  ⑊  rmse.a  rmse.a  rmse.a
--  -  ------  ------  ------
 8  |  0.028   0.0261  0.0253
10  |  0.0290  0.0275  0.0266
"""
gen_test_set(xps_shorter,
             "rmse.a",
             dict(outer=("da_method", "seed"), inner="N"),
             subcols=False)


if "--replace" in sys.argv:
    new_code = orig_code[0:replacements[0].nOpen]

    for i, replacement in enumerate(replacements):

        replacement.lines[0] = 'old = """' + replacement.lines[0]
        new_code += replacement.lines

        try:
            nEnd = replacements[i + 1].nOpen
        except IndexError:
            nEnd = len(orig_code)
        new_code += orig_code[replacement.nClose:nEnd]

    # Write new .new file. Use diff'ing to inspect!
    with open(__file__ + ".new", "w") as F:
        for line in new_code:
            # Remvoe trailing whitespace, which does not get printed
            # by many terminals, and so are not possibly to copy manually.
            # F.write(line.rstrip()+"\n")
            F.write(line)
